@@ -11,6 +11,7 @@ import fnmatch
 import email.utils
 import email.policy
 import email.header
+import email.generator
 import tempfile
 import pathlib
 
@@ -25,7 +26,7 @@ import mailbox
 import pwd
 
 from contextlib import contextmanager
-from typing import Optional, Tuple, Set, List
+from typing import Optional, Tuple, Set, List, TextIO
 
 from email import charset
 charset.add_charset('utf-8', None)
@@ -43,9 +44,22 @@ try:
 except ModuleNotFoundError:
     can_patatt = False
 
-__VERSION__ = '0.7.1'
+__VERSION__ = '0.8.0'
+
+
+def _dkim_log_filter(record):
+    # Hide all dkim logging output in normal operation by setting the level to
+    # DEBUG. If debugging output has been enabled then prefix dkim logging
+    # output to make its origin clear.
+    record.levelno = logging.DEBUG
+    record.levelname = 'DEBUG'
+    record.msg = 'DKIM: ' + record.msg
+    return True
+
 
 logger = logging.getLogger('b4')
+dkimlogger = logger.getChild('dkim')
+dkimlogger.addFilter(_dkim_log_filter)
 
 HUNK_RE = re.compile(r'^@@ -\d+(?:,(\d+))? \+\d+(?:,(\d+))? @@')
 FILENAME_RE = re.compile(r'^(---|\+\+\+) (\S+)')
@@ -57,23 +71,17 @@ ATT_FAIL_FANCY = '\033[31m\u2717\033[0m'
 
 DEVSIG_HDR = 'X-Developer-Signature'
 
-# You can use bash-style globbing here
-WANTHDRS = [
-    'sender',
-    'from',
-    'to',
-    'cc',
-    'subject',
-    'date',
-    'message-id',
-    'resent-message-id',
-    'reply-to',
-    'in-reply-to',
-    'references',
-    'list-id',
-    'errors-to',
-    'x-mailing-list',
-    'resent-to',
+# Headers to include into am-ready messages
+# From: and Subject: are always included
+AMHDRS = [
+    'Date',
+    'Message-Id',
+    'To',
+    'Cc',
+    'Reply-To',
+    'In-Reply-To',
+    'References',
+    'List-Id',
 ]
 
 # You can use bash-style globbing here
@@ -95,6 +103,7 @@ DEFAULT_CONFIG = {
     'midmask': LOREADDR + '/r/%s',
     'linkmask': LOREADDR + '/r/%s',
     'trailer-order': DEFAULT_TRAILER_ORDER,
+    'listid-preference': '*.feeds.kernel.org,*.linux.dev,*.kernel.org,*',
     'save-maildirs': 'no',
     # off: do not bother checking attestation
     # check: print an attaboy when attestation is found
@@ -167,34 +176,7 @@ class LoreMailbox:
             return self.msgid_map[msgid]
         return None
 
-    def backfill(self, revision):
-        if revision in self.covers and self.covers[revision] is not None:
-            patch = self.covers[revision]
-        else:
-            # Find first non-None member in patches
-            lser = self.series[revision]
-            patch = None
-            for patch in lser.patches:
-                if patch is not None:
-                    break
-        logger.info('---')
-        logger.info('Thread incomplete, attempting to backfill')
-        for project in get_lore_projects_from_msg(patch.msg):
-            projurl = 'https://lore.kernel.org/%s/' % project
-            # Try to backfill from that project
-            backfills = get_pi_thread_by_msgid(patch.msgid, useproject=project)
-            if not backfills:
-                continue
-            was = len(self.msgid_map)
-            for msg in backfills:
-                self.add_message(msg)
-            if len(self.msgid_map) > was:
-                logger.info('Loaded %s messages from %s', len(self.msgid_map)-was, projurl)
-            if self.series[revision].complete:
-                logger.info('Successfully backfilled missing patches')
-                break
-
-    def partial_reroll(self, revision, sloppytrailers, backfill):
+    def partial_reroll(self, revision, sloppytrailers):
         # Is it a partial reroll?
         # To qualify for a partial reroll:
         # 1. Needs to be version > 1
@@ -202,7 +184,7 @@ class LoreMailbox:
         if revision <= 1 or revision - 1 not in self.series:
             return
         # Are existing patches replies to previous revisions with the same counter?
-        pser = self.get_series(revision-1, sloppytrailers=sloppytrailers, backfill=backfill)
+        pser = self.get_series(revision-1, sloppytrailers=sloppytrailers)
         lser = self.series[revision]
         sane = True
         for patch in lser.patches:
@@ -256,7 +238,7 @@ class LoreMailbox:
             lser.subject = pser.subject
             logger.debug('Reconstituted successfully')
 
-    def get_series(self, revision=None, sloppytrailers=False, backfill=True, reroll=True):
+    def get_series(self, revision=None, sloppytrailers=False, reroll=True):
         if revision is None:
             if not len(self.series):
                 return None
@@ -278,10 +260,7 @@ class LoreMailbox:
             return None
 
         if not lser.complete and reroll:
-            self.partial_reroll(revision, sloppytrailers, backfill)
-
-        if not lser.complete and backfill:
-            self.backfill(revision)
+            self.partial_reroll(revision, sloppytrailers)
 
         # Grab our cover letter if we have one
         if revision in self.covers:
@@ -442,6 +421,8 @@ class LoreSeries:
         self.has_cover = False
         self.partial_reroll = False
         self.subject = '(untitled)'
+        # Used for base matching
+        self._indexes = None
 
     def __repr__(self):
         out = list()
@@ -598,8 +579,7 @@ class LoreSeries:
                 if noaddtrailers:
                     add_trailers = False
                 msg = lmsg.get_am_message(add_trailers=add_trailers, trailer_order=trailer_order, copyccs=copyccs)
-                slug = '%04d_%s' % (lmsg.counter, re.sub(r'\W+', '_', lmsg.subject).strip('_').lower())
-                msgs.append((slug, msg))
+                msgs.append(msg)
             else:
                 logger.error('  ERROR: missing [%s/%s]!', at, self.expected)
             at += 1
@@ -621,50 +601,109 @@ class LoreSeries:
 
         return msgs
 
-    def check_applies_clean(self, gitdir, when=None):
-        # Go through indexes and see if this series should apply cleanly
-        mismatches = 0
-        seenfiles = set()
-        for lmsg in self.patches[1:]:
-            if lmsg is None or lmsg.blob_indexes is None:
-                continue
-            for fn, bh in lmsg.blob_indexes:
-                if fn in seenfiles:
-                    # if we have seen this file once already, then it's a repeat patch
-                    # and it's no longer going to match current hash
+    def check_applies_clean(self, gitdir: str, at: Optional[str] = None) -> Tuple[int, list]:
+        if self._indexes is None:
+            self._indexes = list()
+            seenfiles = set()
+            for lmsg in self.patches[1:]:
+                if lmsg is None or lmsg.blob_indexes is None:
                     continue
-                seenfiles.add(fn)
-                if set(bh) == {'0'}:
-                    # New file, will for sure apply clean
-                    continue
-                fullpath = os.path.join(gitdir, fn)
-                if when is None:
-                    if not os.path.exists(fullpath):
-                        mismatches += 1
+                for fn, bh in lmsg.blob_indexes:
+                    if fn in seenfiles:
+                        # if we have seen this file once already, then it's a repeat patch
+                        # and it's no longer going to match current hash
                         continue
-                    cmdargs = ['hash-object', fullpath]
-                    ecode, out = git_run_command(None, cmdargs)
-                else:
-                    gitdir = os.path.join(gitdir, '.git')
-                    logger.debug('Checking hash on %s:%s', when, fn)
-                    # XXX: We should probably pipe the two commands instead of reading into memory,
-                    #      so something to consider for the future
-                    ecode, out = git_run_command(gitdir, ['show', f'{when}:{fn}'])
-                    if ecode > 0:
-                        # Couldn't get this file, continue
-                        logger.debug('Could not look up %s:%s', when, fn)
-                        mismatches += 1
+                    seenfiles.add(fn)
+                    if set(bh) == {'0'}:
+                        # New file, will for sure apply clean
                         continue
-                    cmdargs = ['hash-object', '--stdin']
-                    ecode, out = git_run_command(None, cmdargs, stdin=out.encode())
-                if ecode == 0:
-                    if out.find(bh) != 0:
-                        logger.debug('%s hash: %s (expected: %s)', fn, out.strip(), bh)
-                        mismatches += 1
-                    else:
-                        logger.debug('%s hash: matched', fn)
+                    self._indexes.append((fn, bh))
 
-        return len(seenfiles), mismatches
+        mismatches = list()
+        if at is None:
+            at = 'HEAD'
+        for fn, bh in self._indexes:
+            ecode, out = git_run_command(gitdir, ['ls-tree', at, fn])
+            if ecode == 0 and len(out):
+                chunks = out.split()
+                if chunks[2].startswith(bh):
+                    logger.debug('%s hash: matched', fn)
+                    continue
+                else:
+                    logger.debug('%s hash: %s (expected: %s)', fn, chunks[2], bh)
+            else:
+                # Couldn't get this file, continue
+                logger.debug('Could not look up %s:%s', at, fn)
+            mismatches.append((fn, bh))
+
+        return len(self._indexes), mismatches
+
+    def find_base(self, gitdir: str, branches: Optional[str] = None, maxdays: int = 30) -> Tuple[str, len, len]:
+        # Find the date of the first patch we have
+        pdate = datetime.datetime.now()
+        for lmsg in self.patches:
+            if lmsg is None:
+                continue
+            pdate = lmsg.date
+            break
+
+        # Find latest commit on that date
+        guntil = pdate.strftime('%Y-%m-%d')
+        if branches:
+            where = ['--branches', branches]
+        else:
+            where = ['--all']
+
+        gitargs = ['log', '--pretty=oneline', '--until', guntil, '--max-count=1'] + where
+        lines = git_get_command_lines(gitdir, gitargs)
+        if not lines:
+            raise IndexError
+        commit = lines[0].split()[0]
+        checked, mismatches = self.check_applies_clean(gitdir, commit)
+        fewest = len(mismatches)
+        if fewest > 0:
+            since = pdate - datetime.timedelta(days=maxdays)
+            gsince = since.strftime('%Y-%m-%d')
+            logger.debug('Starting --find-object from %s to %s', gsince, guntil)
+            best = commit
+            for fn, bi in mismatches:
+                logger.debug('Finding tree matching %s=%s in %s', fn, bi, where)
+                gitargs = ['log', '--pretty=oneline', '--since', gsince, '--until', guntil,
+                           '--find-object', bi] + where
+                lines = git_get_command_lines(gitdir, gitargs)
+                if not lines:
+                    logger.debug('Could not find object %s in the tree', bi)
+                    continue
+                for line in lines:
+                    commit = line.split()[0]
+                    logger.debug('commit=%s', commit)
+                    # We try both that commit and the one preceding it, in case it was a delete
+                    # Keep track of the fewest mismatches
+                    for tc in [commit, f'{commit}~1']:
+                        sc, sm = self.check_applies_clean(gitdir, tc)
+                        if len(sm) < fewest and len(sm) != sc:
+                            fewest = len(sm)
+                            best = tc
+                            logger.debug('fewest=%s, best=%s', fewest, best)
+                            if fewest == 0:
+                                break
+                        if fewest == 0:
+                            break
+                    if fewest == 0:
+                        break
+                if fewest == 0:
+                    break
+        else:
+            best = commit
+        if fewest == len(self._indexes):
+            # None of the blobs matched
+            raise IndexError
+
+        lines = git_get_command_lines(gitdir, ['describe', '--all', best])
+        if len(lines):
+            return lines[0], len(self._indexes), fewest
+
+        raise IndexError
 
     def make_fake_am_range(self, gitdir):
         start_commit = end_commit = None
@@ -923,13 +962,11 @@ class LoreMessage:
                     self.trailers.append(trailer)
 
     def get_trailers(self, sloppy=False):
-        mismatches = set()
-        if sloppy:
-            return self.trailers, mismatches
-
         trailers = list()
+        mismatches = set()
+
         for tname, tvalue, extdata in self.trailers:
-            if tname.lower() in ('fixes', 'obsoleted-by'):
+            if sloppy or tname.lower() in ('fixes', 'obsoleted-by'):
                 trailers.append((tname, tvalue, extdata, self))
                 continue
 
@@ -1002,16 +1039,17 @@ class LoreMessage:
         # until we come to a passing one
         dkhdrs = list()
         for header in list(self.msg._headers):  # noqa
-            # Also remove any List- headers set by lore.kernel.org
-            if header[0].lower().startswith('list-') and header[1].find('//lore.kernel.org/') > 0:
-                self.msg._headers.remove(header) # noqa
-            elif header[0].lower() == 'dkim-signature':
+            if header[0].lower() == 'dkim-signature':
                 dkhdrs.append(header)
                 self.msg._headers.remove(header) # noqa
         dkhdrs.reverse()
 
         seenatts = list()
         for hn, hval in dkhdrs:
+            # Handle MIME encoded-word syntax or other types of header encoding if
+            # present.
+            if '?q?' in hval:
+                hval = str(email.header.make_header(email.header.decode_header(hval)))
             errors = list()
             hdata = LoreMessage.get_parts_from_header(hval)
             logger.debug('Loading DKIM attestation for d=%s, s=%s', hdata['d'], hdata['s'])
@@ -1030,7 +1068,7 @@ class LoreMessage:
                     signtime = self.date
 
             self.msg._headers.append((hn, hval))  # noqa
-            res = dkim.verify(self.msg.as_bytes())
+            res = dkim.verify(self.msg.as_bytes(), logger=dkimlogger)
 
             attestor = LoreAttestorDKIM(res, identity, signtime, errors)
             logger.debug('DKIM verify results: %s=%s', identity, res)
@@ -1045,6 +1083,27 @@ class LoreMessage:
         # No exact domain matches, so return everything we have
         self._attestors += seenatts
 
+    def _trim_body(self) -> None:
+        # Get the length specified in the X-Developer-Signature header
+        xdsh = self.msg.get('X-Developer-Signature')
+        if not xdsh:
+            return
+        matches = re.search(r'\s+l=(\d+)', xdsh)
+        if not matches:
+            return
+        bl = int(matches.groups()[0])
+        i, m, p = get_mailinfo(self.msg.as_bytes(), scissors=False)
+        bb = b''
+        for line in re.sub(rb'[\r\n]*$', b'', m + p).split(b'\n'):
+            bb += re.sub(rb'[\r\n]*$', b'', line) + b'\r\n'
+        if len(bb) > bl:
+            self.body = bb[:bl].decode()
+            # This may have potentially resulted in in-body From/Subject being removed,
+            # so make sure we account for this in the message headers
+            self.lsubject.subject = self.subject = i.get('Subject')
+            self.fromname = i.get('Author')
+            self.fromemail = i.get('Email')
+
     def _load_patatt_attestors(self) -> None:
         if not can_patatt:
             logger.debug('Message has %s headers, but can_patatt is off', DEVSIG_HDR)
@@ -1066,7 +1125,26 @@ class LoreMessage:
 
         logger.debug('Loading patatt attestations with sources=%s', str(sources))
 
-        attestations = patatt.validate_message(self.msg.as_bytes(), sources)
+        success = False
+        trim_body = False
+        while True:
+            attestations = patatt.validate_message(self.msg.as_bytes(), sources, trim_body=trim_body)
+            # Do we have any successes?
+            for attestation in attestations:
+                if attestation[0] == patatt.RES_VALID:
+                    success = True
+                    break
+            if success:
+                if trim_body:
+                    # If we only succeeded after trimming the body, then we MUST set the body
+                    # to that value, otherwise someone can append arbitrary content after the l= value
+                    # limit message.
+                    self._trim_body()
+                break
+            if not success and trim_body:
+                break
+            trim_body = True
+
         for result, identity, signtime, keysrc, keyalgo, errors in attestations:
             if keysrc and keysrc.startswith('(default keyring)/'):
                 fpr = keysrc.split('/', 1)[1]
@@ -1081,7 +1159,10 @@ class LoreMessage:
                     parts = email.utils.parseaddr(uids[0])
                     identity = parts[1]
 
-            signdt = LoreAttestor.parse_ts(signtime)
+            if signtime:
+                signdt = LoreAttestor.parse_ts(signtime)
+            else:
+                signdt = None
             attestor = LoreAttestorPatatt(result, identity, signdt, keysrc, keyalgo, errors)
             self._attestors.append(attestor)
 
@@ -1105,7 +1186,8 @@ class LoreMessage:
                         # This is not critical even in hardfail
                         continue
                 elif attpolicy in ('softfail', 'hardfail'):
-                    checkmark = attestor.checkmark
+                    if not checkmark:
+                        checkmark = attestor.checkmark
                     trailers.append('%s BADSIG: %s' % (attestor.checkmark, attestor.trailer))
 
                 if attpolicy == 'hardfail':
@@ -1164,7 +1246,7 @@ class LoreMessage:
             if hcs is None:
                 hcs = 'utf-8'
             try:
-                decoded += hstr.decode(hcs)
+                decoded += hstr.decode(hcs, errors='replace')
             except LookupError:
                 # Try as utf-u
                 decoded += hstr.decode('utf-8', errors='replace')
@@ -1194,6 +1276,35 @@ class LoreMessage:
                 msgid = matches.groups()[0]
         return msgid
 
+    @staticmethod
+    def get_preferred_duplicate(msg1, msg2):
+        config = get_main_config()
+        listid1 = LoreMessage.get_clean_msgid(msg1, 'list-id')
+        if listid1:
+            prefidx1 = 0
+            for listglob in config['listid-preference']:
+                if fnmatch.fnmatch(listid1, listglob):
+                    break
+                prefidx1 += 1
+        else:
+            prefidx1 = config['listid-preference'].index('*')
+
+        listid2 = LoreMessage.get_clean_msgid(msg2, 'list-id')
+        if listid2:
+            prefidx2 = 0
+            for listglob in config['listid-preference']:
+                if fnmatch.fnmatch(listid2, listglob):
+                    break
+                prefidx2 += 1
+        else:
+            prefidx2 = config['listid-preference'].index('*')
+
+        if prefidx1 <= prefidx2:
+            logger.debug('Picked duplicate from preferred source: %s', listid1)
+            return msg1
+        logger.debug('Picked duplicate from preferred source: %s', listid2)
+        return msg2
+
     @staticmethod
     def get_patchwork_hash(diff: str) -> str:
         """Generate a hash from a diff. Lifted verbatim from patchwork."""
@@ -1255,8 +1366,11 @@ class LoreMessage:
 
     @staticmethod
     def find_trailers(body, followup=False):
-        headers = ('subject', 'date', 'from')
-        nonperson = ('fixes', 'subject', 'date', 'link', 'buglink', 'obsoleted-by')
+        ignores = {'phone', 'email'}
+        headers = {'subject', 'date', 'from'}
+        nonperson = {'fixes', 'subject', 'date', 'link', 'buglink', 'obsoleted-by'}
+        # Ignore everything below standard email signature marker
+        body = body.split('\n-- \n', 1)[0].strip() + '\n'
         # Fix some more common copypasta trailer wrapping
         # Fixes: abcd0123 (foo bar
         # baz quux)
@@ -1279,6 +1393,9 @@ class LoreMessage:
                 groups = list(matches.groups())
                 # We only accept headers if we haven't seen any non-trailer lines
                 tname = groups[0].lower()
+                if tname in ignores:
+                    logger.debug('Ignoring known non-trailer: %s', line)
+                    continue
                 if len(others) and tname in headers:
                     logger.debug('Ignoring %s (header after other content)', line)
                     continue
@@ -1435,7 +1552,7 @@ class LoreMessage:
             self.body += '\n'
 
         if len(message):
-            self.body += message + '\n'
+            self.body += message.rstrip('\r\n') + '\n'
             if len(fixtrailers):
                 self.body += '\n'
 
@@ -1446,26 +1563,26 @@ class LoreMessage:
                     self.body += '%s\n' % trailer[2]
         if len(basement):
             self.body += '---\n'
-            self.body += basement
-            self.body += '\n'
+            self.body += basement.rstrip('\r\n') + '\n\n'
         if len(signature):
             self.body += '-- \n'
-            self.body += signature
-            self.body += '\n'
+            self.body += signature.rstrip('\r\n') + '\n\n'
 
-    def get_am_subject(self):
+    def get_am_subject(self, indicate_reroll=True):
         # Return a clean patch subject
         parts = ['PATCH']
         if self.lsubject.rfc:
             parts.append('RFC')
-        if not self.revision_inferred:
-            if self.reroll_from_revision:
+        if self.reroll_from_revision:
+            if indicate_reroll:
                 if self.reroll_from_revision != self.revision:
                     parts.append('v%d->v%d' % (self.reroll_from_revision, self.revision))
                 else:
                     parts.append(' %s  v%d' % (' ' * len(str(self.reroll_from_revision)), self.revision))
             else:
                 parts.append('v%d' % self.revision)
+        elif not self.revision_inferred:
+            parts.append('v%d' % self.revision)
         if not self.lsubject.counters_inferred:
             parts.append('%d/%d' % (self.lsubject.counter, self.lsubject.expected))
 
@@ -1474,25 +1591,26 @@ class LoreMessage:
     def get_am_message(self, add_trailers=True, trailer_order=None, copyccs=False):
         if add_trailers:
             self.fix_trailers(trailer_order=trailer_order, copyccs=copyccs)
-        am_body = self.body.rstrip('\r\n')
         am_msg = email.message.EmailMessage()
-        am_msg.set_payload(am_body.encode() + b'\n')
-        # Clean up headers
-        for hdrname, hdrval in self.msg.items():
-            lhdrname = hdrname.lower()
-            wanthdr = False
-            for hdrmatch in WANTHDRS:
-                if fnmatch.fnmatch(lhdrname, hdrmatch):
-                    wanthdr = True
-                    break
-            if wanthdr:
-                new_hdrval = LoreMessage.clean_header(hdrval)
-                # noinspection PyBroadException
-                try:
-                    am_msg.add_header(hdrname, new_hdrval)
-                except:
-                    # A broad except to handle any potential weird header conditions
-                    pass
+        am_msg.set_payload(self.body.encode())
+        am_msg.add_header('Subject', self.get_am_subject(indicate_reroll=False))
+        if self.fromname:
+            am_msg.add_header('From', f'{self.fromname} <{self.fromemail}>')
+        else:
+            am_msg.add_header('From', self.fromemail)
+
+        # Add select headers from the original message
+        for hname in AMHDRS:
+            hval = self.msg.get(hname)
+            if not hval:
+                continue
+            hval = LoreMessage.clean_header(hval)
+            # noinspection PyBroadException
+            try:
+                am_msg.add_header(hname, hval)
+            except:
+                # A broad except to handle any potential weird header conditions
+                pass
         am_msg.set_charset('utf-8')
         return am_msg
 
@@ -1731,7 +1849,7 @@ def git_run_command(gitdir: Optional[str], args: List[str], stdin: Optional[byte
                     logstderr: bool = False) -> Tuple[int, str]:
     cmdargs = ['git', '--no-pager']
     if gitdir:
-        if os.path.isdir(os.path.join(gitdir, '.git')):
+        if os.path.exists(os.path.join(gitdir, '.git')):
             gitdir = os.path.join(gitdir, '.git')
         cmdargs += ['--git-dir', gitdir]
     cmdargs += args
@@ -1845,6 +1963,11 @@ def get_main_config() -> dict:
         # Legacy name was get-lore-mbox, so load those as well
         config = get_config_from_git(r'get-lore-mbox\..*', defaults=config)
         config['trailer-order'] = config['trailer-order'].split(',')
+        config['trailer-order'].remove('*')
+        config['trailer-order'].append('*')
+        config['listid-preference'] = config['listid-preference'].split(',')
+        config['listid-preference'].remove('*')
+        config['listid-preference'].append('*')
         if config['gpgbin'] is None:
             gpgcfg = get_config_from_git(r'gpg\..*', {'program': 'gpg'})
             config['gpgbin'] = gpgcfg['program']
@@ -1952,7 +2075,9 @@ def get_requests_session():
 
 def get_msgid_from_stdin():
     if not sys.stdin.isatty():
-        message = email.message_from_string(sys.stdin.read())
+        from email.parser import BytesParser
+        message = BytesParser().parsebytes(
+            sys.stdin.buffer.read(), headersonly=True)
         return message.get('Message-ID', None)
     return None
 
@@ -2069,26 +2194,33 @@ def get_pi_thread_by_url(t_mbx_url, nocache=False):
         for msg in os.listdir(cachedir):
             with open(os.path.join(cachedir, msg), 'rb') as fh:
                 msgs.append(email.message_from_binary_file(fh))
-        return msgs
+    else:
+        logger.critical('Grabbing thread from %s', t_mbx_url.split('://')[1])
+        session = get_requests_session()
+        resp = session.get(t_mbx_url)
+        if resp.status_code != 200:
+            logger.critical('Server returned an error: %s', resp.status_code)
+            return None
+        t_mbox = gzip.decompress(resp.content)
+        resp.close()
+        if not len(t_mbox):
+            logger.critical('No messages found for that query')
+            return None
+        # Convert into individual files using git-mailsplit
+        with tempfile.TemporaryDirectory(suffix='-mailsplit') as tfd:
+            msgs = mailsplit_bytes(t_mbox, tfd)
+            if os.path.exists(cachedir):
+                shutil.rmtree(cachedir)
+            shutil.copytree(tfd, cachedir)
 
-    logger.critical('Grabbing thread from %s', t_mbx_url.split('://')[1])
-    session = get_requests_session()
-    resp = session.get(t_mbx_url)
-    if resp.status_code != 200:
-        logger.critical('Server returned an error: %s', resp.status_code)
-        return None
-    t_mbox = gzip.decompress(resp.content)
-    resp.close()
-    if not len(t_mbox):
-        logger.critical('No messages found for that query')
-        return None
-    # Convert into individual files using git-mailsplit
-    with tempfile.TemporaryDirectory(suffix='-mailsplit') as tfd:
-        msgs = mailsplit_bytes(t_mbox, tfd)
-        if os.path.exists(cachedir):
-            shutil.rmtree(cachedir)
-        shutil.copytree(tfd, cachedir)
-    return msgs
+    deduped = dict()
+    for msg in msgs:
+        msgid = LoreMessage.get_clean_msgid(msg)
+        if msgid in deduped:
+            deduped[msgid] = LoreMessage.get_preferred_duplicate(deduped[msgid], msg)
+            continue
+        deduped[msgid] = msg
+    return list(deduped.values())
 
 
 def get_pi_thread_by_msgid(msgid, useproject=None, nocache=False, onlymsgids: Optional[set] = None):
@@ -2222,20 +2354,26 @@ def parse_int_range(intrange, upper=None):
             logger.critical('Unknown range value specified: %s', n)
 
 
-def check_gpg_status(status: str) -> Tuple[bool, bool, bool, str, str]:
+def check_gpg_status(status: str) -> Tuple[bool, bool, bool, Optional[str], Optional[str]]:
     good = False
     valid = False
     trusted = False
     keyid = None
-    signtime = ''
+    signtime = None
+
+    # Do we have a BADSIG?
+    bs_matches = re.search(r'^\[GNUPG:] BADSIG ([0-9A-F]+)\s+(.*)$', status, flags=re.M)
+    if bs_matches:
+        keyid = bs_matches.groups()[0]
+        return good, valid, trusted, keyid, signtime
 
     gs_matches = re.search(r'^\[GNUPG:] GOODSIG ([0-9A-F]+)\s+(.*)$', status, flags=re.M)
     if gs_matches:
         good = True
+        keyid = gs_matches.groups()[0]
     vs_matches = re.search(r'^\[GNUPG:] VALIDSIG ([0-9A-F]+) (\d{4}-\d{2}-\d{2}) (\d+)', status, flags=re.M)
     if vs_matches:
         valid = True
-        keyid = vs_matches.groups()[0]
         signtime = vs_matches.groups()[2]
     ts_matches = re.search(r'^\[GNUPG:] TRUST_(FULLY|ULTIMATE)', status, flags=re.M)
     if ts_matches:
@@ -2264,52 +2402,60 @@ def get_gpg_uids(keyid: str) -> list:
     return uids
 
 
-def save_git_am_mbox(msgs: list, dest):
+def save_git_am_mbox(msgs: list, dest: TextIO):
     # Git-am has its own understanding of what "mbox" format is that differs from Python's
     # mboxo implementation. Specifically, it never escapes the ">From " lines found in bodies
     # unless invoked with --patch-format=mboxrd (this is wrong, because ">From " escapes are also
     # required in the original mbox "mboxo" format).
     # So, save in the format that git-am expects
-    # "dest" should be a file handler in writable+binary mode
+    gen = email.generator.Generator(dest, policy=emlpolicy)
     for msg in msgs:
-        bmsg = msg.as_bytes(unixfrom=True, policy=emlpolicy)
-        # public-inbox unixfrom says "mboxrd", so replace it with something else
-        # so there is no confusion as it's NOT mboxrd
-        bmsg = re.sub(b'^From mboxrd@z ', b'From git@z ', bmsg)
-        bmsg = bmsg.rstrip(b'\r\n') + b'\n\n'
-        dest.write(bmsg)
+        msg.set_unixfrom('From git@z Thu Jan  1 00:00:00 1970')
+        gen.flatten(msg, unixfrom=True)
+        gen.write('\n')
 
 
-def get_lore_projects_from_msg(msg) -> list:
-    cachedir = get_cache_dir()
-    listmap = os.path.join(cachedir, 'lists.map.lookup')
-    if not os.path.exists(listmap):
-        # lists.map is a custom service running on lore.kernel.org, so it is
-        # meaningless to make this a configurable URL
-        session = get_requests_session()
-        resp = session.get('https://lore.kernel.org/lists.map')
-        if resp.status_code != 200:
-            logger.debug('Unable to retrieve lore.kernel.org/lists.map')
-            return list()
-        content = resp.content.decode()
-        with open(listmap, 'w') as fh:
-            fh.write(content)
-    else:
-        with open(listmap, 'r') as fh:
-            content = fh.read()
+def save_maildir(msgs: list, dest):
+    d_new = os.path.join(dest, 'new')
+    pathlib.Path(d_new).mkdir(parents=True)
+    d_cur = os.path.join(dest, 'cur')
+    pathlib.Path(d_cur).mkdir(parents=True)
+    d_tmp = os.path.join(dest, 'tmp')
+    pathlib.Path(d_tmp).mkdir(parents=True)
+    for msg in msgs:
+        # make a slug out of it
+        lsubj = LoreSubject(msg.get('subject', ''))
+        slug = '%04d_%s' % (lsubj.counter, re.sub(r'\W+', '_', lsubj.subject).strip('_').lower())
+        with open(os.path.join(d_tmp, f'{slug}.eml'), 'wb') as mfh:
+            mfh.write(msg.as_string(policy=emlpolicy).encode())
+        os.rename(os.path.join(d_tmp, f'{slug}.eml'), os.path.join(d_new, f'{slug}.eml'))
 
-    projmap = dict()
-    for line in content.split('\n'):
-        if line.find(':') <= 0:
-            continue
-        chunks = line.split(':')
-        projmap[chunks[0]] = chunks[1].strip()
 
-    allto = email.utils.getaddresses([str(x) for x in msg.get_all('to', [])])
-    allto += email.utils.getaddresses([str(x) for x in msg.get_all('cc', [])])
-    projects = list()
-    for entry in allto:
-        if entry[1] in projmap:
-            projects.append(projmap[entry[1]])
+def get_mailinfo(bmsg: bytes, scissors: bool = False) -> Tuple[dict, bytes, bytes]:
+    with tempfile.TemporaryDirectory() as tfd:
+        m_out = os.path.join(tfd, 'm')
+        p_out = os.path.join(tfd, 'p')
+        if scissors:
+            cmdargs = ['mailinfo', '--encoding=UTF-8', '--scissors', m_out, p_out]
+        else:
+            cmdargs = ['mailinfo', '--encoding=UTF-8', '--no-scissors', m_out, p_out]
 
-    return projects
+        ecode, info = git_run_command(None, cmdargs, bmsg)
+        if not len(info.strip()):
+            raise ValueError('Could not get mailinfo')
+
+        i = dict()
+        m = b''
+        p = b''
+        for line in info.split('\n'):
+            line = line.strip()
+            if not line:
+                continue
+            chunks = line.split(':',  1)
+            i[chunks[0]] = chunks[1].strip()
+
+            with open(m_out, 'rb') as mfh:
+                m = mfh.read()
+            with open(p_out, 'rb') as pfh:
+                p = pfh.read()
+    return i, m, p
